library KalmanDLL;
uses WinTypes, WinProcs;


     {===================================================|
     |                                                   |
     |  This Kalman Filter was created May 3, 1994 by:   |
     |                                                   |
     |		     Bob Josselson                       |
     |               Alliant Techsystems                 |
     |               Hopkins, MN 55343                   |
     |                                                   |
     |               612-931-5391                        |
     |                                                   |
     |===================================================}



const
  Size            = 10;    {Set Size larger than the number of states}
  SmallDummySize  = Size * Size;
  DummySize       = 2 * SmallDummySize;

type
  Index       		= 0..Size;
  DummyIndex 		= 0..DummySize;
  SmallDummyIndex	= 0..SmallDummySize;
  InVector   		= array [Index]        	   of Double;
  OutVector   		= array [Index]        	   of Double;
  Vector      		= array [Index]        	   of Double;
  DummyVector 		= array [DummyIndex]  	   of Double;
  SmallDummyVector 	= array [SmallDummyIndex]  of Double;
  Matrix      		= array [Index, Index] 	   of Double;
  Global = record
    n  		: double;  {Number of states}
    m  		: double;  {Number of measurements}
    DeltaT	: double;  {Sample Rate}
    PkLow 	: double;  {Lower diagonal state covariance elements}
    PkHigh	: double;  {Upper diagonal state covariance elements}
    QkLow 	: double;  {Lower diagonal process covariance elements}
    QkHigh	: double;  {Upper diagonal process covariance elements}
    Rk00  	: double;  {0,0 element of measurement covariance matrix}
    Rk11  	: double;  {1,1 element of measurement covariance matrix}
    Rk22 	: double;  {2,2 element of measurement covariance matrix}
    xk     	: Vector;  {State vector}
    zk		: Vector;  {Measurement vector}
    Pk 		: Matrix;  {State covariance matrix}
    Kk 		: Matrix;  {Kalman gain matrix}
    Qk 		: Matrix;  {Process covariance matrix}
    Rk 		: Matrix;  {Measurement covariance matrix}
    Fk 		: Matrix;  {State transition matrix}
    Id 		: Matrix;  {Identity matrix}
    Hk 		: Matrix;  {Measurement matrix}
    HkTrans	: Matrix;  {Transpose of measurement matrix}
    Fktrans	: Matrix   {Transpose of state transition matrix}
  end;

{---------------------------------------------------------}
procedure Mident ( n:Index;  var Id:Matrix );
{  form the n x n Identity matrix}
var
  k, l          : integer;

begin
  for k := 0 to n do begin
    for l := 0 to n do begin
      Id[k,l] :=  0;
    end;
    Id[k,k] := 1;
  end;
end; {Mident}

{---------------------------------------------------------}
procedure Madd ( n,m:Index; A,B:Matrix; var Y:Matrix );
{  Y = A + B   form the sum of n x m matrices}
var
  k, l          : integer;

begin
  for k := 0 to n do begin
    for l := 0 to m do begin
      Y[k,l] :=  A[k,l] + B[k,l];
    end;
  end;
end; {Madd}

{---------------------------------------------------------}
procedure Msub ( n,m:Index; A,B:Matrix; var Y:Matrix );
{  Y = A - B   form the difference of n x m matrices}
var
  k, l          : integer;

begin
  for k := 0 to n do begin
    for l := 0 to m do begin
      Y[k,l] :=  A[k,l] - B[k,l];
    end;
  end;
end; {Msub}

{---------------------------------------------------------}
procedure Mtrans ( n, m:Index; Y:Matrix; var YTrans:Matrix );
{  Y = Ytranspose  Y is n x m, Ytranspose is m x n }
var
  k, l          : integer;
  a             : Matrix;

begin
  for k := 0 to n do begin
    for l := 0 to m do  a[k,l] := Y[k,l];
  end;
  for k := 0 to m do begin
    for l := 0 to n do Ytrans[k,l] := a[l,k];
  end;
end; {Mtrans}

{---------------------------------------------------------}
procedure Mmult ( n,m,p:Index; A,B:Matrix; var Y:Matrix );
{  Y = A B   form the product of 2 matrices}
var
  i, k, l       : integer;

begin
  for k := 0 to n do begin
    for l := 0 to p do begin
      Y[k,l] := 0.0;
      for i := 0 to m do Y[k,l] := Y[k,l] + A[k,i] * B[i,l];
    end;
  end;
end; {Mmult}

{---------------------------------------------------------}
procedure Vadd ( n:Index; a, b:Vector; var y:Vector );
{  y = a + b   form the sum of vectors a and b}
var
  k          : integer;

begin
  for k := 0 to n do y[k] := a[k] + b[k];
end; {Vadd}

{---------------------------------------------------------}
procedure Vsub ( n:Index; a, b:Vector; var y:Vector );
{  y = a - b   form the differrence of vectors a and b}
var
  k          : integer;

begin
  for k := 0 to n do y[k] := a[k] - b[k];
end; {Vsub}

{---------------------------------------------------------}
procedure Vmult ( n, m:Index; A:Matrix; b:Vector; var y:Vector );
{  y = A b   form the product of a matrix, A, with a vector, b}
var
  i, k          : integer;

begin
  for k := 0 to n do begin
    y[k] := 0.0;
    for i := 0 to m do begin
      y[k] := y[k] + A[k,i] * b[i];
    end;
  end;
end; {Vmult}

{---------------------------------------------------------}
procedure Minv ( n:Index; x:Matrix; var det:real; var xInv:Matrix );
var
  i,j,ic,ir,n2,nn,ii,irw    : integer;
  irc,irpc,n2ij,iaz,iazup   : integer;
  ielmt,ielmta,icw,ij,jj    : integer;
  jm,k                      : integer;
  aae,aer                   : real;
  aa                        : DummyVector;
  a, aw, aInv               : SmallDummyVector;

begin
      n := n + 1;
      k := 0;
      for i:= 1 to n do begin
        for j := 1 to n do begin
          k := k + 1;
          a[k] := x[j-1,i-1];
        end;
      end;
      det        := 1.0;
      n2 := n + n;
      nn := 2 * n * n;
{ 7}  for i := 1 to nn do aa[i] := 0.0;
      for i := 1 to n do begin
        j := n2 * (i-1) + n + i;
        aa[j] := 1.0;
{ 8}  end; {for}
      for ic := 1 to n do begin
        for ir := 1 to n do begin
          irc := n2 * (ic-1) + ir;
          irpc := n * (ir-1) + ic;
          aa[irc] := a[irpc];
        end;
{ 6}  end;
      for ir := 1 to n do begin {do 4}
        ij := (n2+1) * (ir-1) + 1;
        if abs(aa[ij]) = 0 then begin {else go to 13}
          iaz := 0;
          repeat {do 14}
            iaz := iaz + 1;
            n2ij := (iaz-1) * n2 + ir;
{14}      until ((iaz=n) or ((n2ij >= ij) and (abs(aa[n2ij]) <> 0 )));
{15}      for iazup := 1 to n2 do begin
            ielmt := (ir-1) * n2 + iazup;
            ielmta := (iaz-1) * n2 + iazup;
            aa[ielmt] := aa[ielmta] + aa[ielmt];
{16}      end;
{13}    end; {if}
        aae := aa[ij];
        det := det * aae;
        for ic := 1 to n2 do begin
          ii := n2 * (ir-1) + ic;
          aa[ii] := aa[ii] / aae;
{ 2}    end;
        for ic := 1 to n2 do begin
          ii := n2 * (ir-1) + ic;
          aw[ic] := aa[ii];
{ 3}    end;
        for irw := 1 to n do begin {do 11}
          if irw <> ir then begin {go to 11}
            jj := n2 * (irw-1) + ir;
            aer := aa[jj];
            for icw := 1 to n2 do begin {do 12}
              jm := n2 * (irw-1) + icw;
              aa[jm] := aa[jm] - aw[icw] * aer;
{12}        end; {for}
          end; {if}
{11}    end; {for}
{ 4}  end; {for}
      for ic := 1 to n do begin
        for ir := 1 to n do begin
          irpc := n * (ic-1) + ir;
          irc := n2 * (ir-1) + n + ic;
          aInv[irpc] := aa[irc];
        end; {for]
{15}  end; {for}
      k     := 0;
      for i:= 1 to n do begin
        for j := 1 to n do begin
          k := k + 1;
          xInv[j-1,i-1] := aInv[k];
        end;
      end;
end; {Minv}

{---------------------------------------------------------}
procedure Kalman(var P:Global; var inV:InVector; var outV:OutVector); export;
var
  nn, mm, i, k 	: integer;
  Tk            : Matrix;
  yk            : Vector;
  det		: real;

begin  {Kalman}
  nn := round(P.n) - 1;
  mm := round(P.m) - 1;

        {--  inV[0] = $firstPass  --}
        {--  inV[1] = Pulse Train --}

{--Set Outputs to States on $firstPass}
  if (inV[0] > 0.5) then begin
    for i:= 0 to nn do begin
      outV[i] := P.xk[i];
    end;
  end;

{--Start Main Loop if not $firstPass and Pulse Train > 0.5}
  if (inV[0] < 0.5)  and  (inV[1] > 0.5) then begin

{--Provide Filter Inputs}
    for i:= 2 to (nn+2) do begin
      P.zk[i-2] := InV[i];
    end;

{--Compute Kalman gain }
    Mmult ( nn, nn, mm, P.Pk,      P.HkTrans, Tk   );
    Mmult ( mm, nn, mm, P.Hk,      Tk,        Tk   );
    Madd  ( mm, mm,     Tk,        P.Rk,      Tk   );
    Minv  ( mm,         Tk,        det,       Tk   );
    Mmult ( nn, mm, mm, P.Hktrans, Tk,        Tk   );
    Mmult ( nn, nn, mm, P.Pk,      Tk,        P.Kk ); 

{--Update estimate with measurement }
    Vmult ( mm, nn,  P.Hk,   P.xk,   yk   );
    Vsub  ( mm,      P.zk,   yk,     yk   );
    Vmult ( nn, mm,  P.Kk,   yk,     yk   );
    Vadd  ( nn,      P.xk,   yk,     P.xk );

{--Compute error covariance for updated estimate }
    Mmult ( nn, mm, nn, P.Kk,      P.Hk,      Tk   );
    Msub  ( nn, nn,     P.Id,      Tk,        Tk   );
    Mmult ( nn, nn, nn, Tk,        P.Pk,      P.Pk );   

{--Project Covariance ahead }
    Mmult ( nn, nn, nn,  P.Pk,     P.Fktrans, Tk   );
    Mmult ( nn, nn, nn,  P.Fk,     Tk,        Tk   );
    Madd  ( nn, nn,      Tk,       P.Qk,      P.Pk );

{--Project State ahead }
    Vmult ( nn, nn,  P.Fk,  P.xk,  P.xk );

{--Provide Filter Outputs}
    for i:= 0 to nn do begin
      outV[i] := P.xk[i];
    end;
  end; { Main Loop }
end; {Kalman}

{------------------------------------------------------------------}
function KalmanPA( var pCount:integer):Longint; export;
{Parameter Allocation Function}
begin
  pCount := 10;  {Number of prompter parameters in dialog box}
  KalmanPA :=  sizeof(Global);   {Size of Kalman memory}
end; {KalmanPA}

{------------------------------------------------------------------}
procedure KalmanPI( var P:Global ); export;
{Parameter Initialization Function}
begin
  P.n	   :=  6.0;         {Number of filter states}
  P.m	   :=  3.0;         {Number of measurements}
  P.DeltaT :=  0.1;    	    {Filter Sample Interval; used in Fk Matrix}
  P.PkLow  := 10.0;	    {Pk[0..2] default}
  P.PkHigh := 50.0;	    {Pk[3..5] default}
  P.QkLow  :=  0.1;	    {Qk[0..2] default}
  P.QkHigh :=  0.1;	    {Qk[3..5] default}
  P.Rk00   := 25.0;         {Rk[0,0] default}
  P.Rk11   := 35.0;         {Rk[1,1] default}
  P.Rk22   := 45.0;         {Rk[2,2] default}
end; {KalmanPI}

{------------------------------------------------------------------}
function KalmanPC( var P:Global ):Pchar; export;
{Parameter Change Function}
begin
  KalmanPC :='n;m;DeltaT;PkLow;PkHigh;QkLow;QkHigh;Rk00;Rk11;Rk22';
end; {KalmanPC}

{------------------------------------------------------------------}
procedure KalmanSS( var P:Global; var runCount:longint); export;
{--These are the initial values at the start of simulation}
{--setup required matrices, H, F (Phi), HTrans, FTrans,
       and the identity, Id}
var
  i,j,nn,mm  : integer;
                              
begin
  nn   := round(P.n) - 1;
  mm   := round(P.m) - 1;
  for i := 0 to nn do begin       {Initialize matrix elements to zero}
    for j := 0 to nn do begin
      P.Qk[i,j] := 0.0;
      P.Rk[i,j] := 0.0;
      P.Pk[i,j] := 0.0;
      P.Hk[i,j] := 0.0;
      P.Kk[i,j] := 0.0;
    end;
  end;
  Mident ( nn, P.Id );            {Set the identity matrix}



   {===============================================================|
   |                                                               | 
   |             Fill in your filter matrices below                | 
   |                                                               | 
   |===============================================================}



{=================   Set State Transition Matrix, Fk   =============}
  Mident (nn, P.Fk );
  P.Fk[0,3] := P.DeltaT;
  P.Fk[1,4] := P.DeltaT;
  P.Fk[2,5] := P.DeltaT;
  Mtrans (mm, nn, P.Fk, P.FkTrans);


{=================   Set Measurement Matrix, Hk   ==================}
  Mident (nn, P.Hk );
  Mtrans (mm, nn, P.Hk, P.HkTrans);


{=================   Set State Covariance Matrix, Pk  ==============}
  P.Pk[0,0] :=  P.PkLow;
  P.Pk[1,1] :=  P.PkLow;
  P.Pk[2,2] :=  P.PkLow;
  P.Pk[3,3] :=  P.PkHigh;
  P.Pk[4,4] :=  P.PkHigh;
  P.Pk[5,5] :=  P.PkHigh;

{=================   Set Process Covariance Matrix, Qk  ============}
  P.Qk[0,0] :=  P.QkLow;
  P.Qk[1,1] :=  P.QkLow;
  P.Qk[2,2] :=  P.QkLow;
  P.Qk[3,3] :=  P.QkHigh;
  P.Qk[4,4] :=  P.QkHigh;
  P.Qk[5,5] :=  P.QkHigh;

{=================   Set Measurement Covariance Matrix, Rk  ========}
  P.Rk[0,0] :=  P.Rk00;  
  P.Rk[1,1] :=  P.Rk11;
  P.Rk[2,2] :=  P.Rk22;

{=================   Set Initial States, xk  ========}
  P.xk[0] :=   0.0;
  P.xk[1] :=   0.0;
  P.xk[2] :=   0.0;
  P.xk[3] :=  80.0;
  P.xk[4] := 180.0;
  P.xk[5] := 280.0;

end;  {KalmanSS}

{------------------------------------------------------------------}
procedure KalmanSE( var P:Global; var runCount:longint); export;
{--These are the final values at the end of simulation}

var
  i,j,nn,mm     : integer;
  Outfile	: text;
                              
begin
  assign(Outfile, 'Kalman.dat');
  rewrite(Outfile);
  nn   := round(P.n) - 1;
  mm   := round(P.m) - 1;

  writeln(Outfile, 'Pk');
  for i := 0 to nn do begin
    for j := 0 to nn do write(Outfile, P.Pk[i,j]:10:5);
    writeln(Outfile);
  end;

  writeln(Outfile, 'Kk');
  for i := 0 to nn do begin
    for j := 0 to nn do write(Outfile, P.Kk[i,j]:10:5);
    writeln(Outfile);
  end;

  close(Outfile);
end; {KalmanSE}



exports
  Kalman	index 1,
  KalmanPI	index 4,
  KalmanPA	index 5,
  KalmanPC	index 6,
  KalmanSS	index 7,
  KalmanSE	index 8;

begin
end.

